#ident "$Id: lookup_ldap.c,v 1.6 2004/04/03 07:14:33 raven Exp $"
/*
 * lookup_ldap.c
 *
 *   Copyright 2001-2003 Ian Kent <raven@themaw.net>
 *
 * Module for Linux automountd to access automount maps in LDAP directories.
 *
 */

#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <lber.h>
#include <ldap.h>

#define MODULE_LOOKUP
#include "automount.h"

#define MAPFMT_DEFAULT "sun"

#define MODPREFIX "lookup(ldap): "

/*
 *  Automount entries are stored hierarchically, with the map name as the base
 *  dn for searches on entries for that map.  Thus, to obtain the base dn for
 *  the master map, one would use the following filter:
 *    (&(objectclass=<map_object_class>)(<map_name_attr>="auto.master"))
 *  Once the base dn is obtained (using ldap_get_first_entry, followed by
 *  ldap_get_dn), the following filter will return all entries for the given
 *  map:
 *    (objectclass=<entry_object_class>)
 *  The attributes of interest are <entry_key_attr>, or the key, and
 *  <entry_value_attr> or the value portion of the automount map entry.
 */
struct autofs_schema {
	char *map_object_class;
	char *map_name_attr;

	char *entry_object_class;
	char *entry_key_attr;
	char *entry_value_attr;
};

struct autofs_schema supported_schemas[3] = {
	{ "nisMap", "nisMapName", "nisObject", "cn", "nisMapEntry" },
	{ "automountMap", "ou", "automount", "cn", "automountInformation" },
	{ "automountMap", "automountMapName",
	  "automount", "automountKey", "automountInformation" },
};
#define NR_SCHEMAS	3

struct lookup_context {
	char *server, *base;
	int port;
	/* once we find a schema that works, save it for future lookups */
	struct autofs_schema *schema;
	struct parse_mod *parse;
};

int lookup_version = AUTOFS_LOOKUP_VERSION;	/* Required by protocol */

void set_schema(struct lookup_context *ctxt, struct autofs_schema *schema)
{
	if (!ap.use_old_ldap_lookup)
		ctxt->schema = schema;
}

static LDAP *do_connect(struct lookup_context *ctxt, int *result_ldap)
{
	LDAP *ldap;
	int version = 3;
	int rv;

	if (result_ldap)
		*result_ldap = LDAP_SUCCESS;

	/* Initialize the LDAP context. */
	ldap = ldap_init(ctxt->server, ctxt->port);
	if (!ldap) {
		crit(MODPREFIX "couldn't initialize LDAP connection"
		     " to %s", ctxt->server ? ctxt->server : "default server");
		return NULL;
	}

	/* Use LDAPv3 */
	rv = ldap_set_option(ldap, LDAP_OPT_PROTOCOL_VERSION, &version);
	if (rv != LDAP_SUCCESS) {
		/* fall back to LDAPv2 */
		ldap_unbind(ldap);
		ldap = ldap_init(ctxt->server, ctxt->port);
		if (!ldap) {
			crit(MODPREFIX "couldn't initialize LDAP");
			return NULL;
		} else {
			version = 2;
		}
	}

	/* Connect to the server as an anonymous user. */
	if (version == 2)
		rv = ldap_simple_bind_s(ldap, ctxt->base, NULL);
	else
		rv = ldap_simple_bind_s(ldap, NULL, NULL);

	if (rv != LDAP_SUCCESS) {
		crit(MODPREFIX "couldn't bind to %s",
		     ctxt->server ? ctxt->server : "default server");
		if (result_ldap)
			*result_ldap = rv;
		return NULL;
	}

	return ldap;
}

/*
 *  Returns 1 on success, 0 on failure.
 */
/* XXX need to take into account the BASEDN from the config file */
int autofs_get_basedn_schema(LDAP *ldap, struct lookup_context *ctxt,
			     const char *mapname, struct autofs_schema *schema)
{
	int l, rv;
	char *query, *basedn;
	char *attrs[] = { LDAP_NO_ATTRS, NULL };
	LDAPMessage *result, *e;
	const char *class = schema->map_object_class,
		   *key = schema->map_name_attr;

	/* Build a query string. */
	l = strlen("(&(objectclass=)(=))") + strlen(class) + strlen(key) +
		strlen(mapname) + 1;
	query = alloca(l);
	if (query == NULL) {
		crit(MODPREFIX "malloc: %m");
		return 0;
	}

	memset(query, '\0', l);
	if (sprintf(query, "(&(objectclass=%s)(%s=%s))",
		    class, key, mapname) >= l) {
		debug(MODPREFIX "error forming query string");
		return 0;
	}
	query[l - 1] = '\0';

	/* Look around. */
	debug(MODPREFIX "searching for \"%s\"", query);

	rv = ldap_search_s(ldap, NULL, LDAP_SCOPE_SUBTREE,
			   query, attrs, 0, &result);
	if ((rv != LDAP_SUCCESS) || !result) {
		crit(MODPREFIX "query failed for %s", query);
		return 0;
	}
	e = ldap_first_entry(ldap, result);
	if (!e) {
		debug(MODPREFIX "query succeeded, no matches for %s", query);
		ldap_msgfree(result);
		return 0;
	}

	basedn = ldap_get_dn(ldap, result);
	ldap_msgfree(result);
	if (!basedn)
		return 0;

	ctxt->base = strdup(basedn);
	ldap_memfree(basedn);
	if (!ctxt->base)
		return 0;

	debug("got base dn of %s\n", ctxt->base);
	return 1;
}

int autofs_get_basedn(struct lookup_context *ctxt, const char *mapname)
{
	LDAP *ldap;
	int i, ret;

	ldap = do_connect(ctxt, NULL);
	if (!ldap) {
		error("Unable to bind to the ldap server!\n");
		return -1;
	}

	if (ctxt->schema)
		ret = autofs_get_basedn_schema(ldap, ctxt, mapname,
					       ctxt->schema);
	else {
		for (i = 0; i < NR_SCHEMAS; i++) {
			ret = autofs_get_basedn_schema(ldap, ctxt, mapname,
						       &supported_schemas[i]);
			if (ret == 1) {
				set_schema(ctxt, &supported_schemas[i]);
				break;
			}
		}
	}

	ldap_unbind(ldap);
	return ret;
}

struct lookup_context *context_init(const char *mapname)
{
	struct lookup_context *ctxt = NULL;
	int l;
	char *ptr = NULL;

	ctxt = (struct lookup_context *)malloc(sizeof(struct lookup_context));
	if (ctxt == NULL) {
		crit(MODPREFIX "malloc: %m");
		return NULL;
	}
	memset(ctxt, 0, sizeof(struct lookup_context));

	/* Now we sanity-check by binding to the server temporarily. Be
	 * sure to allow for use of the "default" server, which is set in
	 * an ldap.conf file somewhere. */
	ctxt->server = NULL;
	ctxt->port = LDAP_PORT;
	ctxt->base = NULL;

	ptr = (char *)mapname;
	if (!strncmp(ptr, "//", 2)) {
		char *s = ptr + 2;
		char *p = NULL, *q = NULL;

		/* Isolate the server's name and possibly port. But the :
		   breaks the SUN parser for submounts so we can't actually
		   use it.
		 */
		if ((q = strchr(s, '/'))) {
			if ((p = strchr(s, ':'))) {
				l = p - s;
				p++;
				ctxt->port = atoi(p);
			} else {
				l = q - s;
			}

			ctxt->server = malloc(l + 1);
			memset(ctxt->server, 0, l + 1);
			memcpy(ctxt->server, s, l);

			ptr = q + 1;
		}
	} else if (strchr(ptr, ':') != NULL) {
		l = strchr(ptr, ':') - ptr;
		/* Isolate the server's name. */
		ctxt->server = malloc(l + 1);
		memset(ctxt->server, 0, l + 1);
		memcpy(ctxt->server, mapname, l);
		ptr += l+1;
	}

	/* Isolate the base DN. */
	if (strchr(ptr, '=') == NULL) {
		if (!autofs_get_basedn(ctxt, mapname)) {
			warn(MODPREFIX "Unable to determine base dn\n");
			ctxt->base = NULL;
		}
	}

	if (!ctxt->base) {
		/*
		 *  Previous versions of autofs would simply use
		 *  whatever value was left in 'ptr' as the base dn.  So,
		 *  in order to hopefully not cause regressions, that
		 *  behaviour was preserved in the case where nothing
		 *  more clever could be done.
		 */
		l = strlen(ptr);
		ctxt->base = malloc(l + 1);
		memset(ctxt->base, 0, l + 1);
		memcpy(ctxt->base, ptr, l);
	}

	debug(MODPREFIX "server = \"%s\", port = %d, base dn = \"%s\"",
	      ctxt->server ? ctxt->server : "(default)",
	      ctxt->port, ctxt->base);

	return ctxt;
}

void context_free(struct lookup_context *context)
{
	free(context->server);
	free(context->base);
	free(context);
}

/*
 * This initializes a context (persistent non-global data) for queries to
 * this module.  Return zero if we succeed.
 */
int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **context)
{
	struct lookup_context *ctxt = NULL;
	LDAP *ldap;

	/* If we can't build a context, bail. */
	ctxt = context_init(argv[0]);
	*context = ctxt;
	if (ctxt == NULL)
		return 1;

	/* If a map type isn't explicitly given, parse it like sun entries. */
	if (mapfmt == NULL)
		mapfmt = MAPFMT_DEFAULT;

	/* Initialize the LDAP context. */
	ldap = do_connect(ctxt, NULL);
	if (!ldap)
		return 1;

	/* Okay, we're done here. */
	ldap_unbind(ldap);

	/* Open the parser, if we can. */
	return !(ctxt->parse = open_parse(mapfmt, MODPREFIX, argc-1, argv+1));
}

static int read_one_map(LDAP *ldap, const char *root,
			struct autofs_schema *schema,
			struct lookup_context *ctxt,
			time_t age, int *result_ldap)
{
	int rv, i, j, l, count, keycount;
	char *query;
	LDAPMessage *result, *e;
	char **keyValue = NULL;
	char **values = NULL;
	char *attrs[] = { schema->entry_key_attr,
			  schema->entry_value_attr,
			  NULL };
	const char *class = schema->entry_object_class,
		   *key = schema->entry_key_attr,
		   *type = schema->entry_value_attr;
	int found_entry = 0;

	if (ctxt == NULL) {
		crit(MODPREFIX "context was NULL");
		return 0;
	}

	/* Build a query string. */
	l = strlen("(objectclass=)") + strlen(class) + 1;
	query = alloca(l);
	if (query == NULL) {
		crit(MODPREFIX "malloc: %m");
		return 0;
	}

	memset(query, '\0', l);
	if (sprintf(query, "(objectclass=%s)", class) >= l) {
		debug(MODPREFIX "error forming query string");
	}
	query[l - 1] = '\0';

	/* Look around. */
	debug(MODPREFIX "searching for \"%s\" under \"%s\"",query, ctxt->base);

	rv = ldap_search_s(ldap, ctxt->base, LDAP_SCOPE_SUBTREE,
			   query, attrs, 0, &result);

	if ((rv != LDAP_SUCCESS) || !result) {
		crit(MODPREFIX "query failed for %s", query);
		*result_ldap = rv;
		return 0;
	}

	e = ldap_first_entry(ldap, result);
	if (!e) {
		debug(MODPREFIX "query succeeded, no matches for %s", query);
		ldap_msgfree(result);
		return 0;
	} else
		debug(MODPREFIX "examining entries");

	while (e) {
		keyValue = ldap_get_values(ldap, e, key);

		if (!keyValue || !*keyValue) {
			e = ldap_next_entry(ldap, e);
			continue;
		}

		found_entry = 1;

		values = ldap_get_values(ldap, e, type);
		if (!values) {
			info(MODPREFIX "no %s defined for %s", type, query);
			ldap_value_free(keyValue);
			e = ldap_next_entry(ldap, e);
			continue;
		}

		count = ldap_count_values(values);
		keycount = ldap_count_values(keyValue);
		for (i = 0; i < count; i++) {
			for (j = 0; j < keycount; j++) {
				if (*(keyValue[j]) == '/' &&
				    strlen(keyValue[j]) == 1)
					*(keyValue[j]) = '*';
				cache_add(root, keyValue[j], values[i], age);
			}
		}
		ldap_value_free(values);

		ldap_value_free(keyValue);
		e = ldap_next_entry(ldap, e);
	}

	debug(MODPREFIX "done updating map");

	/* Clean up. */
	ldap_msgfree(result);

	if (found_entry) 
		return 1;
	else
		return 0;
}

static int read_map(const char *root, struct lookup_context *ctxt,
		    time_t age, int *result_ldap)
{
	LDAP *ldap;
	int rv = LDAP_SUCCESS;
	int ret, i;

	/* Initialize the LDAP context. */
	ldap = do_connect(ctxt, &rv);
	if (!ldap) {
		if (rv != LDAP_SUCCESS)
			*result_ldap = rv;
		return 0;
	}

	if (ctxt->schema) {
		ret = read_one_map(ldap, root, ctxt->schema, ctxt, age, &rv);
		if (ret == 1 && rv == LDAP_SUCCESS)
			goto done;
	} else {
		for (i = 0; i < NR_SCHEMAS; i++) {
			ret = read_one_map(ldap, root,
					   &supported_schemas[i],
					   ctxt, age, &rv);
			if (ret == 1 && rv == LDAP_SUCCESS) {
				set_schema(ctxt, &supported_schemas[i]);
				goto done;
			}
		}
	}

	ldap_unbind(ldap);
	if (result_ldap != NULL) 
		*result_ldap = rv;
	return 0;

done:
	/* Clean stale entries from the cache */
	cache_clean(root, age);
	ldap_unbind(ldap);
	return 1;
}

int lookup_ghost(const char *root, int ghost, time_t now, void *context)
{
	struct lookup_context *ctxt = (struct lookup_context *) context;
	struct mapent_cache *me;
	int status = 1, rv = LDAP_SUCCESS;
	char *mapname;
	time_t age = now ? now : time(NULL);

	chdir("/");

	if (!read_map(root, ctxt, age, &rv)) {
		switch (rv) {
		case LDAP_SIZELIMIT_EXCEEDED:
			if (ghost)
				crit("lookup_ghost: Unable to download "
				    "the entire LDAP map for: %s", root);
		case LDAP_UNWILLING_TO_PERFORM:
			return LKP_NOTSUP;
		default:
			return LKP_FAIL;
		}
	}

	if (ctxt->server) {
		int len = strlen(ctxt->server) + strlen(ctxt->base) + 4;

		mapname = alloca(len);
		sprintf(mapname, "//%s/%s", ctxt->server, ctxt->base);
	} else {
		mapname = alloca(strlen(ctxt->base) + 1);
		sprintf(mapname, "%s", ctxt->base);
	}

	status = cache_ghost(root, ghost, mapname, "ldap", ctxt->parse);

	me = cache_lookup_first();
	/* me NULL => empty map */
	if (me == NULL)
		return LKP_EMPTY;

	if (*me->key == '/' && *(root + 1) != '-') {
		me = cache_partial_match(root);
		/* 
		 * me NULL => no entries for this direct mount
		 * root or indirect map
		 */
		if (me == NULL)
			return LKP_EMPTY | LKP_INDIRECT;
	}

	return status;
}

static int lookup_one_schema(LDAP *ldap, const char *root, const char *qKey,
			     struct autofs_schema *schema,
			     struct lookup_context *ctxt)
{
	int rv, i, l, ql;
	time_t age = time(NULL);
	char *query;
	LDAPMessage *result, *e;
	char **values = NULL;
	char *attrs[] = { schema->entry_key_attr,
			  schema->entry_value_attr,
			  NULL };
	const char *class = schema->entry_object_class,
		   *key = schema->entry_key_attr,
		   *type = schema->entry_value_attr;
	struct mapent_cache *me = NULL;
	int ret = CHE_OK;

	if (ctxt == NULL) {
		crit(MODPREFIX "context was NULL");
		return 0;
	}

	/* Build a query string. */
	l = strlen("(&(objectclass=") + strlen(class) + strlen(")");
	l += strlen("(") + strlen(key) + strlen("=") 
				+ strlen(qKey) + strlen("))") + 1;

	query = alloca(l);
	if (query == NULL) {
		crit(MODPREFIX "malloc: %m");
		return 0;
	}

	/* Look around. */
	memset(query, '\0', l);
	ql = sprintf(query, "(&(objectclass=%s)(%s=%s))", class, key, qKey);
	if (ql >= l) {
		debug(MODPREFIX "error forming query string");
		return 0;
	}
	query[l - 1] = '\0';

	debug(MODPREFIX "searching for \"%s\" under \"%s\"", query, ctxt->base);

	rv = ldap_search_s(ldap, ctxt->base, LDAP_SCOPE_SUBTREE,
			   query, attrs, 0, &result);

	if ((rv != LDAP_SUCCESS) || !result) {
		crit(MODPREFIX "query failed for %s", query);
		return 0;
	}

	debug(MODPREFIX "getting first entry for %s=\"%s\"", key, qKey);

	e = ldap_first_entry(ldap, result);
	if (!e) {
		debug(MODPREFIX "got answer, but no first entry for %s", query);
		ldap_msgfree(result);
		return CHE_MISSING;
	}

	debug(MODPREFIX "examining first entry");

	values = ldap_get_values(ldap, e, type);
	if (!values) {
		debug(MODPREFIX "no %s defined for %s", type, query);
		ldap_msgfree(result);
		return 0;
	}

	/* Compare cache entry against LDAP */
	for (i = 0; values[i]; i++) {
		me = cache_lookup(qKey);
		while (me && (strcmp(me->mapent, values[i]) != 0))
			me = cache_lookup_next(me);
		if (!me)
			break;
	}

	if (!me) {
		cache_delete(root, qKey, 0);

		for (i = 0; values[i]; i++) {	
			rv = cache_add(root, qKey, values[i], age);
			if (!rv)
				return 0;
		}

		ret = CHE_UPDATED;
	}

	/* Clean up. */
	ldap_value_free(values);
	ldap_msgfree(result);

	return ret;
}

int lookup_one_connected(LDAP *ldap, const char *root, const char *qKey,
			 struct lookup_context *ctxt)
{
	int ret, i;

	if (ctxt->schema) {
		ret = lookup_one_schema(ldap, root, qKey, ctxt->schema, ctxt);
		debug("lookup_one with schema %s,%s,%s returns %d\n",
		      ctxt->schema->entry_key_attr,
		      ctxt->schema->entry_object_class,
		      ctxt->schema->entry_value_attr, ret);
	} else {
		for (i = 0; i < NR_SCHEMAS; i++) {
			ret = lookup_one_schema(ldap, root, qKey,
					 &supported_schemas[i], ctxt);
			debug("lookup_one with schema %d returns %d\n",i, ret);
			/* If we got a valid result, set the schema. */
			if (ret != CHE_FAIL && !(ret & CHE_MISSING)) {
				set_schema(ctxt, &supported_schemas[i]);
				break;
			}
		}
	}

	return ret;
}

int lookup_one(const char *root, const char *key, int key_len,
	       void *lookup_context)
{
	struct lookup_context *ctxt = (struct lookup_context *)lookup_context;
	LDAP *ldap;
	int ret;

	/* Initialize the LDAP context. */
	ldap = do_connect(ctxt, NULL);
	if (!ldap)
		return CHE_FAIL;

	ret = lookup_one_connected(ldap, root, key, ctxt);
	ldap_unbind(ldap);
	return ret;
}

static int lookup_wild_schema(LDAP *ldap, const char *root,
			      struct autofs_schema *schema,
			      struct lookup_context *ctxt)
{
	int rv, i, l, ql;
	time_t age = time(NULL);
	char *query;
	LDAPMessage *result, *e;
	char **values = NULL;
	char *attrs[] = { schema->entry_key_attr,
			  schema->entry_value_attr,
			  NULL };
	const char *class = schema->entry_object_class,
		   *key = schema->entry_key_attr,
		   *type = schema->entry_value_attr;
	struct mapent_cache *me = NULL;
	int ret = CHE_OK;
	char qKey[KEY_MAX_LEN + 1];
	int qKey_len;

	if (ctxt == NULL) {
		crit(MODPREFIX "context was NULL");
		return 0;
	}

	strcpy(qKey, "/");
	qKey_len = 1;

	/* Build a query string. */
	l = strlen("(&(objectclass=") + strlen(class) + strlen(")");
	l += strlen("(") + strlen(key) + strlen("=") + qKey_len + strlen("))") + 1;

	query = alloca(l);
	if (query == NULL) {
		crit(MODPREFIX "malloc: %m");
		return 0;
	}

	/* Look around. */
	memset(query, '\0', l);
	ql = sprintf(query, "(&(objectclass=%s)(%s=%s))", class, key, qKey);
	if (ql >= l) {
		debug(MODPREFIX "error forming query string");
		return 0;
	}
	query[l - 1] = '\0';

	debug(MODPREFIX "searching for \"%s\" under \"%s\"", query, ctxt->base);
	rv = ldap_search_s(ldap, ctxt->base, LDAP_SCOPE_SUBTREE,
			   query, attrs, 0, &result);

	if ((rv != LDAP_SUCCESS) || !result) {
		crit(MODPREFIX "query failed for %s", query);
		return 0;
	}

	debug(MODPREFIX "getting first entry for %s=\"%s\"", key, qKey);

	e = ldap_first_entry(ldap, result);
	if (!e) {
		debug(MODPREFIX "got answer, but no first entry for %s", query);
		ldap_msgfree(result);
		return 0;
	}

	debug(MODPREFIX "examining first entry");

	values = ldap_get_values(ldap, e, type);
	if (!values) {
		debug(MODPREFIX "no %s defined for %s", type, query);
		ldap_msgfree(result);
		return 0;
	}

	/* Compare cache entry against LDAP */
	for (i = 0; values[i]; i++) {
		me = cache_lookup("*");
		while (me && (strcmp(me->mapent, values[i]) != 0))
			me = cache_lookup_next(me);
		if (!me)
			break;
	}

	if (!me) {
		cache_delete(root, "*", 0);

		for (i = 0; values[i]; i++) {	
			rv = cache_add(root, "*", values[i], age);
			if (!rv)
				return 0;
		}

		ret = CHE_UPDATED;
	}

	/* Clean up. */
	ldap_value_free(values);
	ldap_msgfree(result);

	return ret;
}

static int lookup_wild(LDAP *ldap, const char *root,
		       struct lookup_context *ctxt)
{
	int ret, i;

	if (ctxt->schema)
		return lookup_wild_schema(ldap, root, ctxt->schema, ctxt);


	for (i = 0; i < NR_SCHEMAS; i++) {
		ret = lookup_wild_schema(ldap, root,
					 &supported_schemas[i], ctxt);
		if (ret != CHE_FAIL) {
			set_schema(ctxt, &supported_schemas[i]);
			break;
		}
	}

	return ret;
}

int lookup_mount(const char *root, const char *name, int name_len, void *context)
{
	struct lookup_context *ctxt = (struct lookup_context *) context;
	LDAP *ldap;
	int ret;
	char key[KEY_MAX_LEN + 1];
	int key_len;
	char mapent[MAPENT_MAX_LEN + 1];
	char *mapname;
	struct mapent_cache *me;
	time_t now = time(NULL);
	time_t t_last_read;

	if (ap.type == LKP_DIRECT)
		key_len = snprintf(key, KEY_MAX_LEN, "%s/%s", root, name);
	else
		key_len = snprintf(key, KEY_MAX_LEN, "%s", name);

	if (key_len > KEY_MAX_LEN)
		return 1;

	/* Initialize the LDAP context. */
	ldap = do_connect(ctxt, NULL);
	if (!ldap) {
		warn(MODPREFIX "Unable to connect to the ldap server.  "
		     "Performing mount based on cached data.\n");
		ret = CHE_FAIL;
	} else {
		ret = lookup_one_connected(ldap, root, key, ctxt);
		if (ret == CHE_FAIL) {
			warn(MODPREFIX "Unable to connect to the ldap server."
			     "  Performing mount based on cached data.\n");
		}
	}

	me = cache_lookup_first();
	t_last_read = now - me->age;

	if (ret == CHE_MISSING) {
		if (!cache_delete(root, key, CHE_RMPATH))
			rmdir_path(key);

		/* Have parent update its map */
		if (t_last_read > ap.exp_runfreq)
			kill(getppid(), SIGHUP);

		/* Maybe update wild card map entry */
		if (ap.type == LKP_INDIRECT)
			lookup_wild(ldap, root, ctxt);

	} else if (ret == CHE_UPDATED) {
		/* Have parent update its map */
		if (t_last_read > ap.exp_runfreq)
			kill(getppid(), SIGHUP);
	}
	if (ldap)
		ldap_unbind(ldap);

	me = cache_lookup(key);
	if (me) {
		/* Try each of the LDAP entries in succession. */
		while (me) {
			sprintf(mapent, me->mapent);

			debug(MODPREFIX "%s -> %s", key, mapent);
			ret = ctxt->parse->parse_mount(root, name, name_len,
						  mapent, ctxt->parse->context);
			/* stop after the first successful mount */
			if (ret == 0)
				break;

			/*
			 *  If the mount fails, it is most likely due to
			 *  the directory not existing on the server.  In
			 *  such a case, we want to remember that the mount
			 *  failed so that we don't trigger successive
			 *  mounts.
			 */
			if (ret)
				ret = EXIT_CODE_MOUNT_FAILED;
			me = cache_lookup_next(me);
		}
	} else {
		/* path component, do submount */
		me = cache_partial_match(key);
		if (me) {
			if (ctxt->server) {
				int len = strlen(ctxt->server) +
					    strlen(ctxt->base) + 2 + 1 + 1;
				mapname = alloca(len);
				sprintf(mapname, "//%s/%s", ctxt->server, ctxt->base);
			} else {
				mapname = alloca(strlen(ctxt->base) + 1);
				sprintf(mapname, "%s", ctxt->base);
			}
			sprintf(mapent, "-fstype=autofs ldap:%s", mapname);

			debug(MODPREFIX "%s -> %s", key, mapent);
			ret = ctxt->parse->parse_mount(root, name, name_len,
						  mapent, ctxt->parse->context);
		} else {
			ret = EXIT_CODE_LOOKUP_FAILED;
			error(MODPREFIX "key \"%s\" not found in map.", name);
		}
	}

	return ret;
}

/*
 * This destroys a context for queries to this module.  It releases the parser
 * structure (unloading the module) and frees the memory used by the context.
 */
int lookup_done(void *context)
{
	struct lookup_context *ctxt = (struct lookup_context *) context;
	int rv = close_parse(ctxt->parse);
	free(ctxt->server);
	free(ctxt->base);
	free(ctxt);
	return rv;
}

/*
 *  Check for an entry in an included ldap map.  Named lookup_one_included,
 *  as it is the included map counterpart to lookup_one.
 *
 *  Return Values:
 *    CHE_OK      -  Success
 *    CHE_FAIL    -  malloc failed
 *    CHE_MISSING -  key does not exist in the map
 *    CHE_UPDATED -  either the key was newly added to the cache, or the
 *                   value associated with key changed.
 */
int lookup_one_included(char *map, char *root, char *key, int key_len)
{
	struct lookup_context *ctxt;
	int ret = 0;

	ctxt = context_init(map);
	if (!ctxt)
		return CHE_FAIL;
	ret = lookup_one(root, key, key_len, ctxt);
	context_free(ctxt);
	return ret;
}

/*
 *  Check for the existence of "map", and read in its contents if it
 *  exists.  This is named lookup_map_included as it is the included map
 *  counterpart to read_map.
 *
 *  Return Values:
 *    1 - Success
 *    0 - Failure: malloc failed or there was an ldap failure
 */
int lookup_readmap_included(char *map, char *root, time_t age)
{
	struct lookup_context *ctxt;
	int rv = 0;

	ctxt = context_init(map);
	if (!ctxt)
		return 0;

	/* read_map returns 0 for failure, 1 for success */
	rv = read_map(root, ctxt, age, &rv);

	context_free(ctxt);
	return rv;
}
